/*  XCLIB - a set of C functions for general purposes		      -*- c -*-
    Copyright (C) 1996-2001  Martin Vicente

    This file is part of the eXtended C Library (XCLIB).

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library (see the file COPYING.LIB).
    If not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.		     */


/*|+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++|*\
  |  Martin Vicente, 1996-2001
  |
  | Project : XCLIB (eXtended C Library)
  | Purpose : SYNCHRONOUS I/O MULTIPLEXING
  | Version : 0.3
  | Language: ANSI C (with BSD things)
  | File    : io/io.c
  | Author  : Martin VICENTE <martin@vicente.org>
\*|+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++|*/


#define _XCLIB_INTERNALS

#define _BSD_SOURCE

#define _FACILITY	FACILITY(io)

#include <sys/types.h>
#include <sys/time.h>

#include <errno.h>
#include <stddef.h>
#include <assert.h>

#include <xc/defines.h>
#include <xc/bool.h>
#include <xc/error.h>
#include <xc/xalloc.h>
#include <xc/list.h>

#include "io/ioset.h"

#include "xc/io.h"

#include "global.h"

#define READ	IO_READ
#define WRITE	IO_WRITE
#define EXCEPT	IO_EXCEPT

#define IS_BEFORE_OR_AT(t1, t2)						      \
(SEC(t1) < SEC(t2) || (SEC(t1) == SEC(t2) && USEC(t1) <= USEC(t2)))

#define FIXUP_TIMEVAL(t)						      \
_MBEGIN									      \
    while (USEC(t) <= -MILLION)    { USEC(t) += MILLION; SEC(t)--; }	      \
    if (USEC(t) < 0 && SEC(t) > 0) { USEC(t) += MILLION; SEC(t)--; }	      \
    while (USEC(t) >= +MILLION)    { USEC(t) -= MILLION; SEC(t)++; }	      \
    if (USEC(t) > 0 && SEC(t) < 0) { USEC(t) -= MILLION; SEC(t)++; }	      \
_MEND

#define NONZERO(pt)	(pt == NULL || SEC(*pt) != 0 || USEC(*pt) != 0)

#define SEC(t)	((t).tv_sec)
#define USEC(t)	((t).tv_usec)

typedef struct CheckerRec *Checker;
typedef struct HandlerRec *Handler;
typedef struct TimerRec   *Timer;
typedef struct IdlerRec   *Idler;
typedef struct SignalRec  *Signal;

def_LIST(checkerLIST, __IO, Checker);
def_LIST(handlerLIST, __IO, Handler);
def_LIST(timerLIST,   __IO, Timer);
def_LIST(idlerLIST,   __IO, Idler);
def_LIST(signalLIST,  __IO, Signal);

struct CheckerRec
{
    IOCheckerProc proc;
    IOData       *data;
    IOCheckerId   id;
};

struct HandlerRec
{
    int           fd;
    unsigned int  mask;
    IOHandlerProc proc;
    IOData       *data;
    IOHandlerId   id;
};

struct TimerRec
{
    IOTime        spec;
    IOTimerProc   proc;
    IOData       *data;
    IOTimerId     id;
};

struct IdlerRec
{
    IOCard        count;
    int           freq;
    IOIdlerProc   proc;
    IOData       *data;
    IOIdlerId     id;
};

struct SignalRec
{
    IOSignalProc  proc;
    IOData       *data;
    IOSignalId    id;
};

static void initialize(void);

static bool         loop;
static CheckerList  checkerList;
static HandlerList  handlerList;
static TimerList    timerList;
static IdlerList    idlerList;
static SignalList   signalList;
static struct ioset ioset;

static bool ready = false;

/*****************************************************************************
 *	io_add_checker
 *****************************************************************************/

IOCheckerId
io_add_checker(IOCheckerProc proc, IOData *data)
{
    Checker cc;

    if (!ready) initialize();

    assert(proc != NULL);

    cc = xnew(Checker);

    cc->proc = proc;
    cc->data = data;

    return cc->id = checkerLIST->add(checkerList, cc);
}

/*****************************************************************************
 *	io_del_checker
 *****************************************************************************/

void
io_del_checker(IOCheckerId id)
{
    if (!ready) initialize();

    assert(id != NULL);

#ifdef __DEBUG__
#endif

    xfree(checkerLIST->del(id));
}

/*****************************************************************************
 *	io_add_handler
 *****************************************************************************/

IOHandlerId
io_add_handler(int fd, IOMask mask, IOHandlerProc proc, IOData *data)
{
    Handler hh;

    if (!ready) initialize();

    assert(proc != NULL);

    hh = xnew(Handler);

    hh->fd   = fd;
    hh->mask = mask;
    hh->proc = proc;
    hh->data = data;

    ioset_add(&ioset, fd, mask);

    return hh->id = handlerLIST->append(handlerList, hh);
}

/*****************************************************************************
 *	io_del_handler
 *****************************************************************************/

static int setup_ioset(HandlerId id, void *data);

void
io_del_handler(IOHandlerId id)
{
    if (!ready) initialize();

    assert(id != NULL);

#ifdef __DEBUG__
#endif

    ioset_clr(&ioset);

    handlerLIST->action(handlerList, setup_ioset, NULL);

    xfree(handlerLIST->del(id));
}

static int setup_ioset(HandlerId id, __unused void *data)
{
    Handler const hh = id->item;

    ioset_add(&ioset, hh->fd, hh->mask);

    return 0;
}

/*****************************************************************************
 *	io_add_timer
 *****************************************************************************/

static int test_timer(TimerId id, void *spec);

IOTimerId
io_add_timer(const IOTime *spec, IOTimerProc proc, IOData *data)
{
    Timer tt;

    if (!ready) initialize();

    assert(spec != NULL && proc != NULL);

    tt = xnew(Timer);

    tt->spec = *spec;
    tt->proc =  proc;
    tt->data =  data;

    FIXUP_TIMEVAL(tt->spec);

    {
	TimerId this;

	this = timerLIST->search(timerList, test_timer, &tt->spec);

	if (this != NULL)
	    tt->id = timerLIST->insert_before(tt, this);
	else
	    tt->id = timerLIST->add(timerList, tt);
    }

    return tt->id;
}

static int test_timer(TimerId id, void *_spec)
{
    IOTime spec = *(IOTime *)_spec;

    return !IS_BEFORE_OR_AT(id->item->spec, spec);
}

/*****************************************************************************
 *	io_del_timer
 *****************************************************************************/

void
io_del_timer(IOTimerId id)
{
    if (!ready) initialize();

    assert(id != NULL);

#ifdef __DEBUG__
#endif

    xfree(timerLIST->del(id));
}

/*****************************************************************************
 *	io_add_idler
 *****************************************************************************/

IOIdlerId
io_add_idler(int freq, IOIdlerProc proc, IOData *data)
{
    Idler ii;

    if (!ready) initialize();

    assert(proc != NULL);

    ii = xnew(Idler);

    ii->count = 0;
    ii->freq  = freq;
    ii->proc  = proc;
    ii->data  = data;

    return ii->id = idlerLIST->append(idlerList, ii);
}

/*****************************************************************************
 *	io_del_idler
 *****************************************************************************/

void
io_del_idler(IOIdlerId id)
{
    if (!ready) initialize();

    assert(id != NULL);

#ifdef __DEBUG__
#endif

    xfree(idlerLIST->del(id));
}

/*****************************************************************************
 *	io_add_signal
 *****************************************************************************/

IOSignalId
io_add_signal(IOSignalProc proc, IOData *data)
{
    Signal ss;

    if (!ready) initialize();

    assert(proc != NULL);

    ss = xnew(Signal);

    ss->proc = proc;
    ss->data = data;

    return ss->id = signalLIST->append(signalList, ss);
}

/*****************************************************************************
 *	io_del_signal
 *****************************************************************************/

void
io_del_signal(IOSignalId id)
{
    if (!ready) initialize();

    assert(id != NULL);

#ifdef __DEBUG__
#endif

    xfree(signalLIST->del(id));
}

/*****************************************************************************
 *	io_main_loop
 *****************************************************************************/

static void checker(void);
static int  sleep(IOTime *ptimeout);
static int  timer(void);
static int  idler(void);
static void check_events(fd_set fdset[3]);
static int  call_checker(CheckerId id, void *data);
static int  call_signal(SignalId id, void *data);
static int  test_read_handler(HandlerId id, void *data);
static int  test_write_handler(HandlerId id, void *data);
static int  test_except_handler(HandlerId id, void *data);

void
io_main_loop(void)
{
    if (!ready) initialize();

    while (loop)
    {
	IOTime *ptimeout;

	if (!checkerLIST->is_empty(checkerList)) checker();

	if (!idlerLIST->is_empty(idlerList))
	{
	    IOTime timeout;

	    SEC(timeout)  = 0;
	    USEC(timeout) = 0;

	    ptimeout = &timeout;
	}
	else if (!timerLIST->is_empty(timerList))
	{
	    IOTime timeout;
	    IOTime now;

	    timeout = timerLIST->first(timerList)->item->spec;

	    if (gettimeofday(&now, NULL) == -1)
	    {
		error("gettimeofday() failure");
	    }

	    SEC(timeout)  -= SEC(now);
	    USEC(timeout) -= USEC(now);

	    FIXUP_TIMEVAL(timeout);

	    if (SEC(timeout) < 0 || USEC(timeout) < 0)
	    {
		SEC(timeout)  = 0;
		USEC(timeout) = 0;
	    }

	    ptimeout = &timeout;
	}
	else
	{
	    ptimeout = NULL;
	}

	if ((!handlerLIST->is_empty(handlerList) || NONZERO(ptimeout)) && sleep(ptimeout) > 0) continue;
	if ((!timerLIST  ->is_empty(timerList))                        && timer()         > 0) continue;
	if ((!idlerLIST  ->is_empty(idlerList))                        && idler()         > 0) continue;
    }

    loop = true;
}

static void checker(void)
{
    checkerLIST->action(checkerList, call_checker, NULL);
}

static int call_checker(CheckerId id, __unused void *data)
{
    Checker const cc = id->item;

    cc->proc(cc->id, cc->data);

    return 0;
}

static int sleep(IOTime *ptimeout)
{
    fd_set fdset[3];
    int    ret;

    fdset[READ]   = ioset.fdset[READ];
    fdset[WRITE]  = ioset.fdset[WRITE];
    fdset[EXCEPT] = ioset.fdset[EXCEPT];

    do
    {
	ret = select(
	    ioset.fdmax + 1,
	    &fdset[READ],
	    &fdset[WRITE],
	    &fdset[EXCEPT],
	    ptimeout);

	if (ret == -1 && errno == EINTR)
	{
	    signalLIST->action(signalList, call_signal, NULL);

	    if (!loop) return 1;
	}
    }
    while (ret == -1 && errno == EINTR);

    if (ret == -1) error("select() failure");

    if (ret > 0)
    {
	check_events(fdset);
    }
    else
    {
    }

    return 0;
}

static int call_signal(SignalId id, __unused void *data)
{
    Signal const ss = id->item;

    ss->proc(ss->id, ss->data);

    return 0;
}

static void check_events(fd_set fdset[3])
{
    handlerLIST->action(handlerList, test_except_handler, &fdset[EXCEPT]);
    handlerLIST->action(handlerList, test_write_handler,  &fdset[WRITE]);
    handlerLIST->action(handlerList, test_read_handler,   &fdset[READ]);
}

static int test_read_handler(HandlerId id, void *_fdset)
{
    Handler const        hh    = id->item;
    fd_set  const *const fdset = _fdset;

    if (FD_ISSET(hh->fd, fdset)) hh->proc(hh->id, hh->fd, READ, hh->data);

    return 0;
}

static int test_write_handler(HandlerId id, void *_fdset)
{
    Handler const        hh    = id->item;
    fd_set  const *const fdset = _fdset;

    if (FD_ISSET(hh->fd, fdset)) hh->proc(hh->id, hh->fd, WRITE, hh->data);

    return 0;
}

static int test_except_handler(HandlerId id, void *_fdset)
{
    Handler const        hh    = id->item;
    fd_set  const *const fdset = _fdset;

    if (FD_ISSET(hh->fd, fdset)) hh->proc(hh->id, hh->fd, EXCEPT, hh->data);

    return 0;
}

static int timer(void)
{
    IOTime now;

    assert(timerLIST->first(timerList) != NULL);

    if (gettimeofday(&now, NULL) == -1)
    {
	error("gettimeofday() failure");
    }

    FIXUP_TIMEVAL(now);

    if (IS_BEFORE_OR_AT(timerLIST->first(timerList)->item->spec, now))
    {
	Timer const tt = timerLIST->remove(timerLIST->first(timerList));

	tt->proc(tt->id, &tt->spec, tt->data);

	xfree(tt);

	return 1;
    }

    return 0;
}

static int idler(void)
{
    Idler ii;

    assert(idlerLIST->first(idlerList) != NULL);

    ii = idlerLIST->move_to_end(idlerLIST->first(idlerList))->item;

    ii->proc(ii->id, ii->count++, ii->data);

    return 1;
}

/*****************************************************************************
 *	io_exit_loop
 *****************************************************************************/

void
io_exit_loop(void)
{
    if (!ready) initialize();

    loop = false;
}

/*****************************************************************************
 *	initialize
 *****************************************************************************/

static void initialize(void)
{
    ioset_clr(&ioset);

    checkerList = checkerLIST -> create(0);
    handlerList = handlerLIST -> create(0);
    timerList   = timerLIST   -> create(0);
    idlerList   = idlerLIST   -> create(0);
    signalList  = signalLIST  -> create(0);

    loop = true;

    ready = true;
}


/* End Of File */
